#include <errno.h>

#include "armdefs.h"
#include "armos.h"
#include "armemu.h"

#define EMUL_ADDR  unsigned int

static int parameter_offset;
static ARMul_State * emul_state;

static void trace_ret32(int32_t x);
static void trace_ret32_2(int32_t x1, int32_t x);
static void trace_ret64(int64_t x);

static void
write_status64(uint64_t x)
{
  ARMword cpsr = ARMul_GetCPSR(emul_state);
  if (errno == 0)
    {
      emul_state->Reg[0] = x & 0xffffffff;
      emul_state->Reg[1] = (x >> 32) & 0xffffffff;
      cpsr &= ~CBIT;
    }
  else
    {
      emul_state->Reg[0] = errno;
      cpsr |= CBIT;
    }
  ARMul_SetCPSR(emul_state, cpsr);

  trace_ret64(x);
}


static void
write_status32(int32_t x)
{
  ARMword cpsr = ARMul_GetCPSR(emul_state);
  if (errno == 0)
    {
      emul_state->Reg[0] = x;
      emul_state->Reg[1] = 0;
      cpsr &= ~CBIT;
    }
  else
    {
      emul_state->Reg[0] = errno;
      cpsr |= CBIT;
    }
  ARMul_SetCPSR(emul_state, cpsr);

  trace_ret32(x);
}


static void
write_status32_2(int32_t x1, int32_t x2)
{
  ARMword cpsr = ARMul_GetCPSR(emul_state);
  if (errno == 0)
    {
      emul_state->Reg[0] = x1;
      emul_state->Reg[1] = x2;
      cpsr &= ~CBIT;
    }
  else
    {
      emul_state->Reg[0] = errno;
      cpsr |= CBIT;
    }
  ARMul_SetCPSR(emul_state, cpsr);

  trace_ret32_2(x1, x2);
}


static void
write_memory(EMUL_ADDR dst, const void* src, size_t size)
{
  int i;
  const unsigned char *p = src;
  for (i = 0; i < size; i++)
    ARMul_WriteByte(emul_state, dst + i, *p++);
}


static void
read_memory(void* dst, EMUL_ADDR src, size_t size)
{
  int i;
  unsigned char *p = dst;
  for (i = 0; i < size; i++)
    *p++ = ARMul_ReadByte(emul_state, src + i);
}


#define READ_INT8(EMUL_ADDR) \
    ARMul_ReadByte(emul_state, (EMUL_ADDR))

#define READ_INT16(EMUL_ADDR) \
    ARMul_ReadWord(emul_state, (EMUL_ADDR))

#define READ_INT32(EMUL_ADDR) \
    ARMul_ReadWord(emul_state, (EMUL_ADDR))

#define WRITE_INT8(DST_EMUL_ADDR, VAL) \
    ARMul_WriteByte(emul_state, (DST_EMUL_ADDR), (VAL))

#define WRITE_INT16(DST_EMUL_ADDR, VAL) \
    do { \
      ARMul_WriteByte(emul_state, (DST_EMUL_ADDR) + 1, ((VAL) >> 8) & 0xff); \
      ARMul_WriteByte(emul_state, (DST_EMUL_ADDR)    , ((VAL)     ) & 0xff); \
    } while(0)

#define WRITE_INT32(DST_EMUL_ADDR, VAL) \
    ARMul_WriteWord(emul_state, (DST_EMUL_ADDR), (VAL))

uint16_t
read_int16(EMUL_ADDR src)
{
  uint16_t tmp;
  read_memory(&tmp, src, 2);

  return tmp;
}

void
write_int16(EMUL_ADDR dst, uint16_t src)
{
  uint16_t tmp = src;
  write_memory(dst, &tmp, 2);
}

uint64_t
read_int64(EMUL_ADDR src)
{
  uint64_t tmp;
  read_memory(&tmp, src, 8);

  return tmp;
}

void
write_int64(EMUL_ADDR dst, uint64_t src)
{
  uint64_t tmp = src;
  write_memory(dst, &tmp, 8);
}

#define read_long(X) READ_INT32(X)
#define write_long(X,Y) WRITE_INT32(X, Y)
#define WRITE_LONG(X, Y) WRITE_INT32(X, Y)
#define EMUL_SIZEOF_LONG 4

static uint32_t
get_parameter(int nr)
{
  int real_nr = nr + parameter_offset;

  if (real_nr < 4)
    return emul_state->Reg[real_nr];
  else
    return READ_INT32(emul_state->Reg[13] + (real_nr - 4) * 4);
}

#include "../netbsd/netbsd.c"
